; -----------------------------------------------------------------------------
; ColorPRINT-48
; version 1.0
; Copyright (c) 2009 Andrew Owen

; -----------------------------------------------------------------------------
; ColorPRINT-48 is a MODE 2 characater printing tool for the ZX Spectrum 48K.
; It incorporates joefish's 'Rainbow Generator for 48K Spectrum using PUSH'.
; joefish's code turns half of the Spectrum screen into a display equivalent to
; Timex MODE 2. The MODE 2 area is centered on the screen and the rest of the
; display is mostly unaffected. Anything that switches off interrupts, such as
; a BEEP command, will interrupt the display.
;
; The code is only designed to work with 48K machines so it assumes the
; original ROM is in place and uses the empty bytes at 0x3bff as an interrupt
; vector table. It will still work on 128 models in 48 mode without additional
; hardware connected as even the +3 preserves the first two bytes of the table
; at this address.
;
; ColorPRINT-48 uses channels/streams to extend 48 BASIC.
; Use it with PRINT #4 CHR$ n (where n is 0 to 255). The co-ordinates are taken
; from the standard screen channel. To print at a specific co-ordiante use
; PRINT AT y,x; #4; "some text";. All 256 characters can be redefined. USR "A"
; is the first byte of the bitmap of character 144. USR "A" + 2048 is the first
; byte of the attributes of character 144. In the font area, USR "A" - 376 is
; the start of the bitmap of 'a', while USR "A" - 632 is 'A' and so on (the
; CHARS and UDG system variables are changed to point at the new data area).
; PRINT, PLOT, CIRCLE and DRAW now function on the bitmap only and will not
; modify the attributes in the MODE 2 screen area.

; -----------------------------------------------------------------------------
; ROM ROUTINES, SYSTEM VARIABLES and CONSTANTS

	pr_all		equ	0x0b7f	; print all routine
	po_store	equ	0x0adc	; position store routine
	report_j	equ	0x15c4	; null input routine
	make_room	equ	0x1655	; create buffer routine
	strms		equ	0x5c10	; pointers to individual streams
	chars		equ	0x5c36	; start of font - 256
	chans		equ	0x5c4f	; start of channel area
	prog		equ	0x5c53	; start of BASIC area
	udg		equ	0x5c7b	; start of UDGs
	df_cc		equ	0x5c84	; display file addres
	s_posn		equ	0x5c88	; line/column
	attr_p		equ	0x5c8d	; permanent attribute

; -----------------------------------------------------------------------------
; USER ROUTINES

; select MODE 0
	org	52000
mode_0:
	im	1
	ret

; select MODE 2
	org	52003
mode_2:
	im	2
	ret

; cls
	org	52006
cls:
	call	0x0d6b			; clear bitmap
	ld	a, (attr_p)		; get permanent attribute value in A
	ld	hl, colour_map		; set attributes
	ld	de, colour_map + 1	;
	ld	(hl), a			; set first byte
	ld	bc, 3071		;
	ldir				; set remaining 3071 bytes
	ret				; done
	
; -----------------------------------------------------------------------------
; INITIALIZE

	org	52025
start:

; save registers
	push	af
	push	bc
	push	de 
	push	hl

; -----------------------------------------------------------------------------
; MAKE CHANNEL and ATTACH STREAM
; Based on code by Ian Beardsmore from Your Spectrum #7, September 1984.

make_chan:
	ld	hl, (prog)	; a channel must be created below BASIC
				; so look at the system variable PROG
	dec	hl		; move HL down one address
	ld	bc, 0x0005	; the new channel takes 5 bytes
	call	make_room	; call the MAKE_ROOM routine
	inc	hl		; move HL up one address
	ld	bc, c_print	; could write the bytes directly but
 				; then code would be non-relocatable
	ld	(hl), c		; low byte of the output routine
	inc	hl		; move HL up one address
	push	hl		; save this address for later
	ld	(hl), b		; high byte of the output routine
	inc	hl		; move HL up one address
	ld	bc, report_j	; address of input routine
	ld	(hl), c		; low byte of the input routine
	inc	hl		; move HL up one address
	ld	(hl), b		; high byte of the input routine
	inc	hl		; move HL up one address
	ld	(hl), 'S'	; channel type: 'S' = screen

; attach stream
	pop	hl		; the first address plus one of the
				; extra space stored earlier
	ld	de, (chans)	; store the contents of CHANS in DE
	and	a		; clear the carry flag before
				; calculation
	sbc	hl,de		; the difference between the start of
				; the channels area and the start of the
				; extra space becomes the offset, stored
				; in HL
	ex	de, hl		; store the offset in DE
	ld	hl, strms	; store the contents of STRMS in HL
	ld	a, 4		; stream number #4
	add	a, 3		; take account of streams -3 to -1
	add	a, a		; each of the seven default streams has
				; two bytes of offset data
				; the total number of bytes occupied,
				; held in a, forms the offset for the
				; new stream
	ld	b, 0x00		; set b to hold 0x00
	ld	c,a		; set the low byte of the offset
	add	hl,bc		; the offset is added to the base
				; address to give the correct location
				; in the streams table to store the
				; offset
	ld	(hl),e		; the low byte of the offset
	inc	hl		; move HL up one address
	ld	(hl),d		; the high byte of the offset

; -----------------------------------------------------------------------------

; set CHARS and UDG

set_chars_udg:
	ld	hl, chars	; address system variable
	ld	(hl), 0x00	; set
	inc	hl		; font
	ld	(hl), 0xcc	; address
	ld	hl, udg		; address system variable
	ld	(hl), 0x80	; set
	inc	hl		; udg
	ld	(hl), 0xd0	; address

; clear the screen and the attribute area
	call	cls		; clear the attribute data

; install interrupts
	ld	a,0x3b		; High byte of address table...
	ld	i,a		; ...goes into the interrupt vector
	im	2		; Select interrupt mode 2 (custom)

; restore registers

	pop	hl
	pop	de
	pop	bc
	pop	af

; return to BASIC
	ret 

; -----------------------------------------------------------------------------
; DATA

; udg data
; --------
	org	52224

font_address:
	incbin	"ColorFONT.bin"

; attribute map
; -------------
	org	56320
colour_map:

; -----------------------------------------------------------------------------
; SERVICE ROUTINES

; color print channel
; -------------------
	org	59392
c_print:
	ld	bc, (chars)	; font address
	ld	h, 0		; set high byte to 0
	ld	l, a		; character to L
	add	hl, hl		; multiply by 8 to point to first byte
	add	hl, hl		;
	add	hl, hl		;
	add	hl, bc		;
	ex	de, hl		; character address to DE
	ld	bc, (s_posn)	; get line/column
	push	bc		; save line/column
	ld	hl, (df_cc)	; get display file address
	push	de		; store character address
	call	pr_all		; pr-all
	call	po_store	; po-store
	pop	de		; get character address
	ex	de, hl		; store it in HL
	ld	bc, 2048	; offset to character attributes
	add	hl, bc		; add offset
	ex	de, hl		; character attribute address to DE

; error trapping doesn't try to print attributes
; outside the rainbow area
	pop	bc		; get line/column
	ld	a, c		; get column
	cp	10		; column 24 or higher
	ret	c		;
	cp	26		; column 7 or lower
	ret	nc		;

; calculate attribute addres from s_posn
	ld	a, 24		; actual line
	sub	b		; to a
	ld	h, 0		; clear h
	ld	l, a		; low byte to l

	add	hl, hl		; x2
	add	hl, hl		; x4
	add	hl, hl		; x8
	add	hl, hl		; x16
	add	hl, hl		; x32
	add	hl, hl		; x64
	add	hl, hl		; x128

	ld	a, 25		; actual column
	sub	c		; to a
		
	ld	bc, colour_map	;
	add	a, c		;
	ld	c, a		;
	
	add	hl, bc		; hl now points to colour_map

	ld	c, 8		; loop counter
loop:
	ld	a, (de)		; get the value
	ld	(hl), a		; write it
 
 	push	de
	ld	de, 16
	add	hl, de
	pop	de
 
	inc	de		; next attribute byte
	dec	c		; next counter
	jr	nz, loop	; finished?
	ret			; return

; main interrupt routine [from a 2009 WoS thread by joefish]
; ----------------------------------------------------------
	org	60095
interrupt_main:

; Save all registers, then save the Stack Pointer:
	push	af
	push	bc
	push	de
	push	hl
	ex	af,af'
	exx
	push	af
	push	bc
	push	de
	push	hl
	push	ix
	push	iy
	ld	(sp_cache),sp

;Pause for the top border:
	ld	bc, 435
	
interrupt_delay_loop:
	nop				; 04t => 004t
	nop				; 04t => 008t
	dec	bc			; 06t => 014t
	ld	a,b			; 04t => 018t
	or	c			; 04t => 022t
	jp	nz,interrupt_delay_loop	; 10t => 032t per loop,
					; 7 loops per 224t line

	nop				; timing adjustment tweaks
	nop
	nop

; copy the colours
REPT 24,char_row		; PASMO Macro for 24 rows
REPT 8,pixel_row		; PASMO Macro for 8 lines per row
	ld	sp,colour_map+(16*pixel_row)+(128*char_row)	; set the stack pointer to the data
	pop	af		; scoop up 8 bytes into registers
	pop	bc
	pop	de
	pop	hl
	ex	af,af'		; exchange registers for alternates
	exx
	pop	af		; scoop up 8 more bytes into registers
	pop	bc
	pop	de
	pop	hl
	ld	sp,22552+(32*char_row)	; set the stack pointer to the screen
	push	hl		; write 8 bytes from registers
	push	de
	push	bc
	push	af
	exx			; exchange registers for alternates
	ex	af,af'
	push	hl		; write 8 more bytes from registers
	push	de
	push	bc
	push	af
	nop			; timing adjustment tweaks
	nop
ENDM				; repeat 8 lines using the same screen address
ENDM				; repeat 24 rows

; restore the stack pointer, then all the registers:
	ld	sp,(sp_cache)
	pop	iy
	pop	ix
	pop	hl
	pop	de
	pop	bc
	pop	af
	exx
	ex	af,af'
	pop	hl
	pop	de
	pop	bc
	pop	af

; exit via standard interrupt handler
; (so that the keyboard still works in BASIC)
	jp	56

; im2 jump to routine
	org	0xfff4
	jp	interrupt_main

; cache for the stack pointer
sp_cache
	defw	0

; im2 entry point
	org	0xffff
	defb	0x18

; end of program
; --------------
; set entry point for PASMO's
; .TAP file generator:
	end	52000
